// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/geom/sector_2d.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_cross_to_center.h"
#include "body_kick_to_corner.h"

#include "bhv_keep_shoot_chance.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_KeepShootChance::execute( rcsc::PlayerAgent * agent )
{
    static const double goalie_max_speed = 0.95;

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_KeepShootChance"
                        ,__FILE__, __LINE__ );

    rcsc::Vector2D pass_point;
    if ( rcsc::Body_Pass::get_best_pass( agent->world(),
                                         &pass_point,
                                         NULL,
                                         NULL ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:  pass found (%.1f, %.1f)"
                            ,__FILE__,
                            pass_point.x, pass_point.y );
        if ( pass_point.x > 30.0
             && pass_point.absY() < rcsc::ServerParam::i().goalHalfWidth() + 4.0 )
        {
            agent->debugClient().addMessage( "KeepChancePass" );
            rcsc::Body_Pass().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            return true;
        }
    }

    const rcsc::PlayerObject * opp_goalie = agent->world().getOpponentGoalie();
    rcsc::Vector2D goalie_pos( 100.0, 0.0 );
    double goalie_dist = 15.0;
    int goalie_conf = 0;

    if ( opp_goalie )
    {
        goalie_pos = opp_goalie->pos();
        goalie_dist = opp_goalie->distFromSelf();
        goalie_conf = std::min( 5, opp_goalie->posCount() );
    }

    // goalie is not close
    if ( goalie_dist + goalie_max_speed * goalie_conf > 6.5
         && goalie_pos.x - agent->world().self().pos().x > 6.5
         )
    {
        if ( agent->world().self().pos().x > 36.0 //42.0
             && Bhv_CrossToCenter::get_best_point( agent ) )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: cross"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "KeepChanceCross" );
            Bhv_CrossToCenter().execute( agent );
            return true;
        }

#if 1
        if ( agent->world().getOpponentsFromSelf().empty()
             || agent->world().getOpponentsFromSelf().front()->distFromSelf()
             > rcsc::ServerParam::i().defaultPlayerSpeedMax()
             + rcsc::ServerParam::i().defaultKickableArea()
             + 0.8 )
        {
            rcsc::Vector2D drib_target( rcsc::ServerParam::i().pitchHalfLength(),
                                        rcsc::ServerParam::i().goalHalfWidth() + 2.0 );
            if ( agent->world().self().pos().y < 0.0 )
            {
                drib_target.y *= -1.0;
            }
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: can turn dribble to goal post"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "KeepChanceDrib" );
            // no dodge
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower() * 0.5,
                                1, // only dash
                                false
                                ).execute( agent );

            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            return true;
        }
#endif

        // dribble to my body dir
        if ( agent->world().self().pos().x < rcsc::ServerParam::i().pitchHalfLength() - 5.0
             && agent->world().self().pos().absY() < rcsc::ServerParam::i().goalHalfWidth() + 10.0
             && agent->world().self().body().abs() < 50.0 )
        {
            rcsc::Vector2D drib_target = agent->world().self().pos();
            drib_target += rcsc::Vector2D::polar2vector(5.0, agent->world().self().body());
            rcsc::Sector2D sector( agent->world().self().pos(),
                                   0.5, 3.0,
                                   agent->world().self().body() - 30.0,
                                   agent->world().self().body() + 30.0 );
            // opponent check with goalie
            if ( ! agent->world().existOpponentIn( sector, 10, true ) )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: dribble to my body"
                                    ,__FILE__, __LINE__ );
                agent->debugClient().addMessage( "KeepChanceDribToBody" );
                // no dodge
                rcsc::Body_Dribble( drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower(),
                                    2,
                                    false
                                    ).execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
                return true;
            }
        }

        // dribble to goal line
        rcsc::Sector2D front_sector( agent->world().self().pos(),
                                     0.5, 6.0,
                                     -30.0, 30.0 );
        // opponent check with goalie
        if ( ! agent->world().existOpponentIn( front_sector, 10, true ) )
        {
            rcsc::Vector2D drib_target( //agent->world().self().pos().x,
                                       rcsc::ServerParam::i().pitchHalfLength(),
                                       rcsc::ServerParam::i().goalHalfWidth() + 2.0 );
            if ( agent->world().self().pos().y < 0.0 )
            {
                drib_target.y *= -1.0;
            }
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: dribble to goal post"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "KeepChanceDribToGoal" );
            // no dodge
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                2,
                                false
                                ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            return true;
        }

        if ( Bhv_CrossToCenter::get_best_point( agent ) )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: cross"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "KeepChanceCross(2)" );
            Bhv_CrossToCenter().execute( agent );
            return true;
        }

        // kick to far side
        if ( agent->world().self().pos().x > 40.0
             && agent->world().self().pos().absY() < 15.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: goal near. kick to far side"
                                ,__FILE__, __LINE__ );
            // to far side
            agent->debugClient().addMessage( "KeepChanceToFar(1)" );
            Body_KickToCorner( (agent->world().self().pos().y > 0.0)
                               ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return true;
        }

        // pass
        if ( rcsc::Body_Pass().execute( agent ) )
        {
            agent->debugClient().addMessage( "KeepChancePass(2)" );
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
            return true;
        }
    }

    if ( Bhv_CrossToCenter::get_best_point( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: cross"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "KeepChanceCross(3)" );
        Bhv_CrossToCenter().execute( agent );
        return true;
    }

    // kick to far side
    if ( agent->world().self().pos().x > 40.0
         && agent->world().self().pos().absY() < 15.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: goal near. kick to far side"
                            ,__FILE__, __LINE__ );
        // to far side
        agent->debugClient().addMessage( "KeepChanceToFar(2)" );
        Body_KickToCorner( (agent->world().self().pos().y > 0.0) ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    // keep
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 200.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );
    if ( closest_opp_dist > 5.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: keep"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "KeepChanceDribCenter" );
        rcsc::Vector2D drib_target( 45.0, 0.0 );
        rcsc::Body_Dribble( drib_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower() * 0.6,
                            1
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    if ( rcsc::Body_Pass().execute( agent ) )
    {
        agent->debugClient().addMessage( "KeepChancePass(3)" );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    // cross far side
    if ( agent->world().self().pos().absY() < 10.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: enforce far side"
                            ,__FILE__, __LINE__ );
        // to far side
        agent->debugClient().addMessage( "KeepChanceToFar(2)" );
        Body_KickToCorner( (agent->world().self().pos().y > 0.0) ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: cross far side"
                        ,__FILE__, __LINE__ );

    // to far side
    agent->debugClient().addMessage( "KeepChanceToFar(3)" );
    Body_KickToCorner( (agent->world().self().pos().y > 0.0) ).execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
    return true;
}
